﻿//#define IS_ECHOING

//**********************************************************************
//  DLMS/COSEM components
//
//  i-cube, 1376 Goumoens-la-Ville, Switzerland
//          infos@icube.ch
//
// (C)Copyright 2004-2017
//**********************************************************************

using System;
using System.Collections.Generic;
using System.Text;

using System.IO.Ports;

namespace ICube.DemoCOM
{
    using ICube.EzHdlc;

    /// <summary>
    /// This class shows how to subclass System.IO.Stream to suppress echoing
    /// from a SerialPort.
    /// It is made specifically for EzHdlc.HDLCProtocol to send/receive data 
    /// to/from the Meter.
    /// Only <b>int ReadByte()</b> and <b>void Write(byte[] buffer, int offset, int count)</b>
    /// are implemented. The other methods throw NotImplementedException.
    /// </summary>
    class COMStream : System.IO.Stream
    {
        public System.IO.Stream stream_;
        private bool IsEchoing;

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="stream">SerialPort Stream to send/receive data.</param>
        /// <param name="IsEchoing">Whether echoing is supresses or not.</param>
        public COMStream(System.IO.Stream stream, bool IsEchoing)
        {
            this.stream_ = stream;
            this.IsEchoing = IsEchoing;
        }

        /// <summary>
        /// Read next byte from the stream.
        /// </summary>
        /// <returns></returns>
        public override int ReadByte()
        {
            return this.stream_.ReadByte();
        }

        /// <summary>
        /// Write to the stream.
        /// If IsEchoing is true, reads count' bytes from the stream
        /// to suppress echoing.
        /// </summary>
        /// <param name="buffer"></param>
        /// <param name="offset"></param>
        /// <param name="count"></param>
        public override void Write(byte[] buffer, int offset, int count)
        {
            this.stream_.Write(buffer, offset, count);

            if (this.IsEchoing)
            {
                // get count bytes from the stream to swallow the echo
                // this could be done as follow
                for (int i = 0; i < count; i++)
                    this.stream_.ReadByte();
            }
        }

        public override bool CanRead
        {
            get { throw new NotImplementedException(); }
        }

        public override bool CanSeek
        {
            get { throw new NotImplementedException(); }
        }

        public override bool CanWrite
        {
            get { throw new NotImplementedException(); }
        }

        public override void Flush()
        {
            throw new NotImplementedException();
        }

        public override long Length
        {
            get { throw new NotImplementedException(); }
        }

        public override long Position
        {
            get
            {
                throw new NotImplementedException();
            }
            set
            {
                throw new NotImplementedException();
            }
        }

        public override int Read(byte[] buffer, int offset, int count)
        {
            throw new NotImplementedException();
        }

        public override long Seek(long offset, System.IO.SeekOrigin origin)
        {
            throw new NotImplementedException();
        }

        public override void SetLength(long value)
        {
            throw new NotImplementedException();
        }

    }

    /// <summary>
    /// 
    /// </summary>
    class DemoCOMSuppressEchoing
    {
        private COMStream comStream;

        private SerialPort serialPort_;
#if IS_ECHOING
        static private bool ISECHOING = true;
#else
        static private bool ISECHOING = false;
#endif
        private ICube.EzHdlc.HDLCParameters hdlcParameters;
        private HDLCProtocol hdlc;
        private DateTime dateTime_ = DateTime.Now;

        DemoCOMSuppressEchoing()
        {
            this.hdlcParameters = new HDLCParameters();
            this.hdlcParameters.SentStreamCallback = this.OnSentStream;
            this.hdlcParameters.ReceivedStreamCallBack = this.OnReceivedStream;
            this.hdlcParameters.StateChangeCallBack = this.OnStateChange;
            this.hdlcParameters.ClientMACAddress = 0x10;		// Public client
            this.hdlcParameters.ServerAddressSize = 1;
            this.hdlcParameters.ServerUpperMACAddressValue = 1;
            this.hdlcParameters.ServerLowerMACAddressValue = 0;
        }

        private void doIt()
        {
            // Open SerialPort
            this.serialPort_ = new SerialPort("COM1");
            this.serialPort_.ReadTimeout = 3000;
            this.serialPort_.BaudRate = 9600;
            this.serialPort_.Parity = System.IO.Ports.Parity.None;
            this.serialPort_.DataBits = 8;
            this.serialPort_.StopBits = System.IO.Ports.StopBits.One;
            this.serialPort_.DtrEnable = true;
            this.serialPort_.Open();

            // Create COMStream
            this.comStream = new COMStream(this.serialPort_.BaseStream, ISECHOING);

            // Creat HDLCProtocol
            this.hdlc = new HDLCProtocol(this.hdlcParameters, this.comStream);
            {
                this.hdlc.Connect();
                Console.WriteLine("  Connected to: COM1");

                Console.WriteLine("Disconnecting");
                this.hdlc.Disconnect();
            }
        }

        protected void OnSentStream(object stream)
        {
            if (stream.GetType() == typeof(byte[]))
                this.LogLine("CB_S: " + ICube.EzHdlc.Utils.BytesToHexString((byte[])stream));
            else
                this.LogLine("CB_S: " + (string)stream);
        }

        protected void OnReceivedStream(object stream)
        {
            if (stream.GetType() == typeof(byte[]))
                this.LogLine("CB_R: " + ICube.EzHdlc.Utils.BytesToHexString((byte[])stream));
            else
                this.LogLine("CB_R: " + (string)stream);
        }

        protected void OnStateChange(HDLCProtocolState state)
        {
            this.LogLine("CB_s: " + state.ToString());
        }

        protected void LogLine(string log)
        {
            TimeSpan _timeSpan = DateTime.Now - this.dateTime_;
            string _timeTag = _timeSpan.TotalSeconds.ToString("F3");
            Console.WriteLine("{0,9} : {1}", _timeTag, log);
        }
        static void Main(string[] args)
        {
            DemoCOMSuppressEchoing demo = new DemoCOMSuppressEchoing();
            demo.doIt();
#if(DEBUG)
            Console.WriteLine("Done hit any key !!!");
            Console.ReadKey();
#endif
        }

    }


}


