﻿//**********************************************************************
//  DLMS/COSEM components
//
//  i-cube, 1376 Goumoens-la-Ville, Switzerland
//          infos@icube.ch
//
// (C)Copyright 2004-2017
//**********************************************************************

using System;
using System.Collections.Generic;
using System.Text;

using System.Threading;
using System.IO;
using System.IO.Ports;
namespace ICube.HdlcDemo
{
    using ICube.EzHdlc;

    using ICube.XmlPdu;
    using ICube.XmlPdu.Infra;

    /// <summary>
    /// This demo shows how to communicate with a Meter using the Hdlc communication
    /// profile and a SerialPort.
    /// 
    /// The meter's HDLC parameters are defined in InitializeParameters().
    /// The serial port parameters are defined in doIt().
    /// 
    /// It processes all xml files given as arguments as follow:
    /// - convert to pdu
    /// - send pdu and receive it's response
    /// - convert pdu response to xml.
    /// </summary>
    class HdlcDemo
    {
        private SerialPort serialPort_;
        protected ICube.EzHdlc.HDLCParameters hdlcParameters;
        protected ICube.EzHdlc.HDLCProtocol hdlc = null;
        protected DateTime dateTime_ = DateTime.Now;

        /// <summary>
        /// Initializes parametres structure.
        /// <remarks>Adapt to your meter parameters.</remarks>
        /// <param name="filename"></param>
        /// </summary>
        protected void InitializeParameters()
        {
            this.hdlcParameters = new HDLCParameters();
            this.hdlcParameters.SentStreamCallback = this.OnSentStream;
            this.hdlcParameters.ReceivedStreamCallBack = this.OnReceivedStream;
            this.hdlcParameters.StateChangeCallBack = this.OnStateChange;

            // LG
            this.hdlcParameters.ClientMACAddress = 0x10;		// Public client
            this.hdlcParameters.ServerAddressSize = 1;        //
            this.hdlcParameters.ServerUpperMACAddressValue = 1;
            this.hdlcParameters.ServerLowerMACAddressValue = 0;

            //// Actaris SL-7000
            //this.hdlcParameters.ClientMACAddress = 0x10;		// Public client
            ////this.hdlcParameters.ClientMACAddress = 0x03;		// Non-Public client
            //this.hdlcParameters.ServerAddressSize = 4;          // 
            //this.hdlcParameters.ServerUpperMACAddressValue = 1;
            //this.hdlcParameters.ServerLowerMACAddressValue = 17;

        }

        /// <summary>
        /// Processes an xml file.
        /// </summary>
        /// <param name="filePath"></param>
        protected virtual bool ProcessXmlFile(string filePath)
        {
            bool _success = false;
            StringBuilder _outXml, _inXml;
            MemoryStream _outPdu, _inPdu;

            Console.WriteLine("--------        --------        --------        --------        --------");
            Console.WriteLine("Processing: " + filePath);

            try
            {
                _outXml = new StringBuilder(File.ReadAllText(filePath));
                Console.Write(_outXml.ToString());
                Console.WriteLine();
                _outPdu = XmlPduInterface.XmlToPdu(_outXml);

                _inPdu = this.hdlc.SendReceive(_outPdu.ToArray());
                _inXml = XmlPduInterface.PduToXml(_inPdu);

                Console.Write(_inXml.ToString());
                _success = true;
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
            }
            Console.WriteLine();
            return _success;
        }

        public void doIt(string[] args)
        {
            Console.WriteLine("HdlcDemo");
            Console.WriteLine(ICube.EzHdlc.Version.VersionStr);
            Console.WriteLine(ICube.XmlPdu.Version.VersionStr);
            Console.WriteLine(ICube.Wrapper.Version.VersionStr);

            this.InitializeParameters();

            // Open SerialPort
            this.serialPort_ = new SerialPort("COM12");
            this.serialPort_.ReadTimeout = 3000;    // NOTE: it is necessary to define a timeout
            this.serialPort_.BaudRate = 9600;
            this.serialPort_.Parity = System.IO.Ports.Parity.None;
            this.serialPort_.DataBits = 8;
            this.serialPort_.StopBits = System.IO.Ports.StopBits.One;
            this.serialPort_.DtrEnable = true;
            this.serialPort_.Open();

            this.hdlc = new HDLCProtocol(this.hdlcParameters, this.serialPort_.BaseStream);
            {
                this.hdlc.Connect();
                Console.WriteLine("  Connected to: " + this.serialPort_.PortName);

                // Process the following xml files
                this.ProcessXmlFile("..\\..\\xml-files\\" + "AssociationRequest_LN_NoPW.xml");
                this.ProcessXmlFile("..\\..\\xml-files\\" + "GetRequestNormal(Clock).xml");

                Console.WriteLine("Disconnecting");
                this.hdlc.Disconnect();
            }
        }

        static void Main(string[] args)
        {
            try {

                HdlcDemo test = new HdlcDemo();
                test.doIt(args);

            }
            catch (EzHdlcException e)
            {
                Console.WriteLine("EZHdlc exception: Kind-{0} {1}", e.Kind, e.Message);

            }
            catch (XmlPduException e)
            {
                Console.WriteLine("XmlPduException exception: {0}", e.Message);

            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);

            }
            Console.WriteLine("Done hit any key !!!");
            Console.ReadKey();
        }

        protected void OnSentStream(object stream)
        {
            this.dateTime_ = DateTime.Now;
            if (stream.GetType() == typeof(byte[]))
                this.Log("CB_S: " + ICube.EzHdlc.Utils.BytesToHexString((byte[])stream));
            else
                this.Log("CB_S: " + (string)stream);
        }

        protected void OnReceivedStream(object stream)
        {
            TimeSpan timeSpan = DateTime.Now - this.dateTime_;
            if (stream.GetType() == typeof(byte[]))
                this.Log(timeSpan, "CB_R: " + ICube.EzHdlc.Utils.BytesToHexString((byte[])stream));
            else
                this.Log(timeSpan, "CB_R: " + (string)stream);
        }

        protected void OnStateChange(HDLCProtocolState state)
        {
            this.Log("CB_s: " + state.ToString());
        }

        protected void Log(string log)
        {
            Console.WriteLine("{0,8} {1,5} {2}", DateTime.Now.ToString("T"), "", log);
        }
        protected void Log(TimeSpan timeSpan, string log)
        {
            Console.WriteLine("{0,8} {1,5} {2}", DateTime.Now.ToString("T"), timeSpan.TotalSeconds.ToString("F3"), log);
        }

    }

}
