﻿//**********************************************************************
//  DLMS/COSEM components
//
//  10-dec-09
//      Added added SetBaudDelay processing in Open()
//
//  i-cube, 1376 Goumoens-la-Ville, Switzerland
//          infos@icube.ch
//
// (C)Copyright 2004-2017
//**********************************************************************

using System;
using System.Collections.Generic;
using System.Text;

using System.IO;
using System.IO.Ports;
using System.Threading;

namespace ICube.Demo
{
    using ICube.EzHdlc;

    /// <summary>
    /// Structure IEC parameters.
    /// </summary>
    public struct IECParameters
    {
        /// <summary>Opening baudrate.</summary>
        public int OpeningBaud;

        /// <summary>Max baudrate.</summary>
        public int MaxBaud;

        /// <summary>Swap baud delay.</summary>
        public int SwapBaudDelay;

        /// <summary>Device address.</summary>
        public string ModeEDeviceAddress;

        public bool IsEchoing;
    }

    public class IECException : Exception
    {
        public IECException(String message)
            : base(message)
        {
        }
    }

    public class IECOpening
    {

        #region Constants

        const byte CR = 0x0d;
        const byte LF = 0x0a;
        const byte ACK = 0x06;

        #endregion

        #region Properties

        public Stream IOStream
        {
            get { return this.serialPort_.BaseStream; }
        }

        ///// <summary>
        ///// Callback called when a frame is sent to the server. 
        ///// </summary>
        //public SentReceivedStreamDelegate OnSentStream { get; set; }

        ///// <summary>
        ///// Callback called when a frame is received from the server.
        ///// </summary>
        //public SentReceivedStreamDelegate OnReceivedStream { get; set; }

        #endregion

        #region Private data members

        private static string[] nonPrintableASCIIs = new string[] { 
                "NUL",  "SOH",  "STX",  "ETX",  "EOT",  "ENQ",  "ACK",  "BEL",
                "BS",   "HT",   "LF",   "VT",   "FF",   "CR",   "SO",   "SI",
                "DLE",  "DC1",  "DC2",  "DC3",  "DC4",  "NAK",  "SYN",  "ETB",
                "CAN",  "EM",   "SUB",  "ESC",  "FS",   "GS",   "RS",   "US" };

        private SerialPort serialPort_;
        private IECParameters iecParameters_;
        #endregion

        public IECOpening(string portName, IECParameters iecParameters)
        {
            try
            {
                this.serialPort_ = new SerialPort(portName);
                this.iecParameters_ = iecParameters;
            }
            catch (Exception e)
            {
                ThrowIECException(e.Message);
            }
        }

        internal void Open()
        {
            int[] _charToBaud = new int[] { 300, 600, 1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200 };
            byte[] _ackString = { ACK, (byte)'2', 0, (byte)'2', CR, LF };
            bool _timeout = false;
            byte _baudChar;
            int baud;
            string s;

            this.serialPort_.Open();
            this.serialPort_.ReadTimeout = 3000;
            // Configure port
            this.Purge();
            this.Configure(iecParameters_.OpeningBaud, System.IO.Ports.Parity.Even, 7,
                System.IO.Ports.StopBits.One);

            // Send identification string
            s = "/?" + iecParameters_.ModeEDeviceAddress + "!\r\n";

            Console.WriteLine("--> {0}", GetReadableASCII(s));

            Write(Encoding.ASCII.GetBytes(s));

            // Get and check response
            MemoryStream _res = GetResponse(LF, ref _timeout);
            byte[] _resBytes = _res.ToArray();
            if (_timeout)
                ThrowIECException("No response to mode E opening");

            if ((_resBytes.Length < 6) || (_resBytes[0] != '/') ||
                !((_resBytes[4] >= '0') && (_resBytes[4] <= '9')))
                ThrowIECException("Mode E illegal response");

            _baudChar = _resBytes[4];
            // Determine requested baud
            do
            {
                baud = _charToBaud[_baudChar - (byte)'0'];
                if ((_baudChar == '0') || (baud <= iecParameters_.MaxBaud))
                    break;
                _baudChar--;
            } while (true);

            // Acknowledge
            if (iecParameters_.SwapBaudDelay > 0)
                Thread.Sleep(iecParameters_.SwapBaudDelay);
            _ackString[2] = _baudChar;
            Console.WriteLine("--> {0}", GetReadableASCII(_ackString));
            Write(_ackString);
            //TBC FlushFileBuffers(comChannel->hCom);
            // Instead of waiting for a response and (sometimes) having a collision
            // between the response and the baud changeover, we simply wait a while,
            // purge the comm channel and then set the new channel parameters.
            Thread.Sleep(2000);
            Purge();
            this.Configure(baud, System.IO.Ports.Parity.None, 8,
                System.IO.Ports.StopBits.One);
            Console.WriteLine("--- baud changeover --- {0}", baud);
            Thread.Sleep(20);
        }


        private MemoryStream GetResponse(byte lastChar, ref bool timeout)
        {
            MemoryStream _result = new MemoryStream();
            byte _readByte;
            bool _done = false;
            do
            {
                _readByte = GetByte(ref timeout);
                if (!timeout)
                {
                    _result.WriteByte(_readByte);
                    _done = (_readByte == lastChar);
                }
            } while ((!_done) && (!timeout));
            if (_result.Length > 0)
            {
                Console.WriteLine("<-- {0}", GetReadableASCII(_result.ToArray()));
            }
            return _result;
        }

        private static string GetReadableASCII(string s)
        {
            return GetReadableASCII(Encoding.ASCII.GetBytes(s));
        }

        private static string GetReadableASCII(byte[] stream)
        {
            string _result = "";
            
            for (int _i = 0; _i < stream.Length; _i++)
            {
                if (stream[_i] < nonPrintableASCIIs.Length)
                    _result += "<" + nonPrintableASCIIs[stream[_i]] + ">";
                else
                    _result += Encoding.ASCII.GetString(stream, _i, 1);
            }

            return _result;
        }

        private void Configure(int baudRate, Parity parity, int dataBits, StopBits stopBits)
        {
            try
            {
                this.serialPort_.BaudRate = baudRate;
                this.serialPort_.Parity = parity;
                this.serialPort_.DataBits = dataBits;
                this.serialPort_.StopBits = stopBits;
                this.serialPort_.DtrEnable = true; 
            }
            catch (Exception e)
            {
                ThrowIECException(e.Message);
            }
        }

        private void Purge()
        {
            try
            {
                this.serialPort_.DiscardInBuffer();
                this.serialPort_.DiscardOutBuffer();
            }
            catch (Exception e)
            {
                ThrowIECException(e.Message);
            }
        }

        public void Write(byte[] buffer)
        {
            try
            {
                int _bufferLength = buffer.Length;
                //aru this.serialPort_.DiscardOutBuffer();
                this.serialPort_.Write(buffer, 0, _bufferLength);

                if (iecParameters_.IsEchoing)
                {
                    bool _timeout = false;
                    int _bytesRead = 0;
                    byte _byte = 0;

#if (DEBUG)
                    byte[] _echoBytes = new byte[_bufferLength];
                    int _echoPos = 0;
#endif
                    while ((_bytesRead != _bufferLength) && !_timeout)
                    {
                        _byte = (byte)this.GetByte(ref _timeout);
                        _bytesRead++;
#if (DEBUG)
                        _echoBytes[_echoPos] = _byte;
                        _echoPos++;
#endif
                    }
                }
            }
            catch (Exception e)
            {
                ThrowIECException(e.Message);
            }
        }

        public byte GetByte(ref bool hasTimedOut)
        {
            byte _readByte = 0;
            bool _timeoutRaised = false;

            try
            {
                _readByte = (byte)this.serialPort_.ReadByte();
            }
            catch (TimeoutException)
            {
                _timeoutRaised = true;
            }
            catch (Exception e)
            {

                ThrowIECException(e.Message);
            }

            hasTimedOut = _timeoutRaised;
            return _readByte;
        }

        private void ThrowIECException(string message)
        {
            throw new IECException("IEC-opening: " + message);
        }

    }
}
